import nltk
from nltk.corpus import treebank
import re
# from nltk.tag import StanfordPOSTagger
# from nltk.tag import SennaTagger
from textblob import TextBlob
import inflect
import re
import ast
import sys
from dateutil.parser import parse

HTTP_verbs = {
'post': ["create", "add", "produce", "make", "put", "write", "pay", "send", "build", 
"raise", "develop", "register", "post", "submit", "apply", "order", "insert","comment"],

'get': ["retrieve", "check", "choose", "request", "search", "contact", "get", 
"take", "see", "ask", "show", "watch", "read", "open", "reach", "return",
"receive", "view", "load", "review", "select"],

'put': ["perform", "mark", "evaluate", "update", "set", "change", "edit"],

'delete': ["delete", "destroy", "kill", "remove","cancel"]}

# response code lists
L404 = {
    "en" : ['not found','doesn\'t exist','does not exist','unable to find','can\'t find','no', 'didn\'t find', 'did not find', 'none found', 'none was found'],
    "el" : ['Δεν βρέθηκε', 'Δεν βρήκαμε', 'Μη υπαρκτό']
    }
L401 = {
    "en" : ['unauthorized'],
    "el" : ['δεν είναι συνδεδεμένος', 'δεν είναι συνδεδεμένη', 'δεν έχει συνδεθεί']
    }
L403 = {
    "en" : ['not allowed', 'rejected', 'denied', 'not approved', 'forbidden'],
    "el" : ['δεν επιτρέπεται', 'απορρίπτεται', 'απαγορεύεται', 'μη εξουσιοδοτημένο', 'μη εξουσιοδοτημένη', 'μη εξουσιοδοτημένος', 'μη επιτρεπτό']    
}
L400 = {
    "en" : ['failed','unsuccessful','wrong','mistake','error','not'],
    "el" : ['λανθασμένο', 'ελλειπές', 'πρόβλημα', 'λάθος']
    }
L200 = {
    "en" : ['success','ok','updated','deleted','retrieved','canceled'],
    "el" : ['επιτυχές', 'επιτυχής', 'επιτυχία', 'ok', 'οκ', 'διαγράφηκε', 'ανακτήθηκε', 'ακυρώθηκε']
    }
L201 = {
    "en" : ['created'],
    "el" : ['δημηουργήθηκε']
    }
L202 = {
    "en" : ['accepted'],
    "el" : ['αποδεκτό']
    }

# phrase lists
maximum_phrases = ['max', 'maximum', 'cannot be more than',\
 'can\'t be more than', 'ceiling', 'can\'t be greater than', 'cannot be greater than']
minimum_phrases = ['min', 'minimum', 'cannot be less than', \
'can\'t be less than', 'floor', 'can\'t be less than', 'cannot be less than']
exclusive_phrases = ['exclusive','including']

# st = StanfordPOSTagger("/path/to/stanfordtagger")

# senna_tagger = SennaTagger("path/to/sennatagger")
p = inflect.engine()

def resource_analysis(resources,resource_names):
    model = {}
    resource_graph = {}
    state_graph = {}
    table_parameter_occurence = {}
    plain_parameter_occurence = {}
    put_non_body_par_occurence = {}
    operation_occurence = {}
    for resource,scenarios in resources.items():
        resource_graph[resource] = []
        model[resource]= {'get':{'request_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}}, 'description':''},
        'post':{'request_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}},'description':''},
        'put':{'request_params':[],'non_body_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}},'description':''},
        'delete':{'request_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}},'description':''},'resource_hierarchy':{},'resource_roles':{}}
        table_parameter_occurence[resource] = {}
        plain_parameter_occurence[resource] = {}
        put_non_body_par_occurence[resource] = {}
        operation_occurence[resource] = {}
        resource_roles = []
        #----------------- Analyze Given steps in Background and scenarios
        for scenario,steps_types in scenarios.items():
            if scenario == 'background':
                for steps_type,steps in steps_types.items():
                     if steps_type == 'Given':
                        for step in steps:
                            resource_hierarchy = detect_other_resources(step['sentence'],resource,resource_names)
                            r_roles = detect_roles(step['sentence'],resource_names,resource_roles)
                            if resource_hierarchy:
                                hierarchy_parameter = detect_parameters(step['sentence'],resource_names)
                                if len(hierarchy_parameter) == 1:
                                    model[resource]['resource_hierarchy'] = {'hierarchy_resource':resource_hierarchy, 'linking_parameter':hierarchy_parameter[0]}
                                else:
                                    sys.exit("Please specify one and only one hierarchy linking parameter between the resources '" + resource + "' and '" + resource_hierarchy + "' in the sentence: "\
                                        + "'"+step['sentence']+"'")
                            elif r_roles:
                                for r_role in r_roles:
                                    resource_roles.append(r_role)
                                    model[resource]['resource_roles'][r_role] = []
        for scenario,steps_types in scenarios.items():                            
            if scenario != 'background':
                when_ops = []
                scenario_roles = []
                for steps_type,steps in steps_types.items():
                    if steps_type == 'Given':    
                        for step in steps:
                            s_roles = detect_roles(step['sentence'],resource_names,resource_roles)
                            if s_roles:
                                for s_role in s_roles:
                                    if s_role not in scenario_roles:
                                        scenario_roles.append(s_role)
        #----------------- Analyze When & Then Steps                            
                for steps_type,steps in steps_types.items():
                    if steps_type == 'When': # 'When' comes always before 'Then', because the dictionary is ordered from the preprocessor
                        plain_when_params = []
                        for step in steps:
                            if 'When' in step['sentence']:
                                plain_when_params = detect_parameters(step['sentence'],resource_names)
                                when_ops = detect_operations(step['sentence'],resource_names,steps_type)
                        for op in when_ops:
                            if op not in operation_occurence[resource]:
                                operation_occurence[resource][op] = {'request_count':1}
                            else:
                                operation_occurence[resource][op]['request_count'] += 1
                        for step in steps:
                            if step['data_table']:
                                processed_data_table = {}  
                                if ('query' in step['sentence'] and len(when_ops)==1 and when_ops[0]=='put'):                           
                                    processed_data_table = analyze_data_table(step['data_table'],'put_non_body')
                                    if 'put' not in operation_occurence[resource].keys():
                                        put_non_body_par_occurence[resource]['put'] = {'request': {}}
                                    elif 'request' not in operation_occurence[resource]['put']:
                                        put_non_body_par_occurence[resource]['put']['request'] = {}
                                    if not processed_data_table['domain'] in model[resource]['put']['non_body_params']:
                                        model[resource]['put']['non_body_params'].append(processed_data_table['domain'])
                                        for param in processed_data_table['domain']:
                                            if not param['name'] in put_non_body_par_occurence[resource]['put']['request'].keys():
                                                put_non_body_par_occurence[resource]['put']['request'][param['name']] = {'count':1}
                                            else:
                                                put_non_body_par_occurence[resource]['put']['request'][param['name']]['count'] += 1                
                                elif step['sentence'].endswith(':'):
                                    for op in when_ops:
                                        processed_data_table = analyze_data_table(step['data_table'],op)
                                        if (op == 'post' or op == 'put'):
                                            model_name = detect_model_name(step['sentence'])
                                            if not model_name:
                                                model_name = 'Uknown_Model'
                                            if not any(model_name == param['name'] for param in model[resource][op]['request_params']):
                                                if processed_data_table['example']:
                                                    model[resource][op]['request_params'].append({'name':model_name, 'model_type':True, 'properties':processed_data_table['domain'], 'example':processed_data_table['example']})
                                                else:
                                                    model[resource][op]['request_params'].append({'name':model_name, 'model_type':True, 'properties':processed_data_table['domain']})
                                            if op not in plain_parameter_occurence[resource].keys():
                                                plain_parameter_occurence[resource][op] = {'request': {}}
                                            if not model_name in plain_parameter_occurence[resource][op]['request']:
                                                plain_parameter_occurence[resource][op]['request'][model_name] = {'count':1}
                                            else:
                                                plain_parameter_occurence[resource][op]['request'][model_name]['count'] += 1
                                else:
                                    for op in when_ops:
                                        processed_data_table = analyze_data_table(step['data_table'],op)
                                        if op not in table_parameter_occurence[resource]:
                                            table_parameter_occurence[resource][op] = {'request': {}}
                                        elif 'request' not in table_parameter_occurence[resource][op]:
                                            table_parameter_occurence[resource][op]['request'] = {}
                                        if not processed_data_table['domain'] in model[resource][op]['request_params']:
                                            model[resource][op]['request_params'].append(processed_data_table['domain'])
                                            if processed_data_table['example']:
                                                model[resource][op]['request_examples'].append(processed_data_table['example'])
                                            for param in processed_data_table['domain']:
                                                if not param['name'] in table_parameter_occurence[resource][op]['request'].keys():
                                                    table_parameter_occurence[resource][op]['request'][param['name']] = {'count':1}
                                                else:
                                                    table_parameter_occurence[resource][op]['request'][param['name']]['count'] += 1
                                    if ('When' in step['sentence']):
                                        if len(when_ops) == 1 and (when_ops[0] == 'put' or when_ops[0] == 'delete' or when_ops[0] == 'get'):
                                            if len(plain_when_params) == 1:
                                                if len(processed_data_table['domain']) == 1:
                                                    for i, table in enumerate(model[resource][when_ops[0]]['request_params']):
                                                        if type(table) is list:
                                                            if len(table) == 1:
                                                                for j, param in enumerate(table):
                                                                    if param['name'] != plain_when_params[0] and param == processed_data_table['domain'][0]:
                                                                        print(param['name'],plain_when_params[0],processed_data_table['domain'],resource)
                                                                        sys.exit('In resource file ' + resource + ' and operation ' + when_ops[0] + ' a single parameter in the sentence and a single parameter \
                                                                            in the data table were specified with different names, exiting')
                                                                    elif param['name'] == plain_when_params[0] and param == processed_data_table['domain'][0]:
                                                                        model[resource][when_ops[0]]['request_params'][i][j]['path_param'] = True
                            else:                           
                                for op in when_ops: 
                                    if op not in plain_parameter_occurence:
                                        plain_parameter_occurence[resource][op] = {'request': {}}
                                    elif 'request' not in plain_parameter_occurence[resource][op]:
                                        plain_parameter_occurence[resource][op]['request'] = {}
                                    if (op == 'put' or op == 'get' or op == 'delete') and ('When' in step['sentence']) and (len(plain_when_params) == 1 or len(plain_when_params) == 0):
                                        if len(plain_when_params) == 1:
                                            param = plain_when_params[0]
                                            duplicate_path_param = False
                                            for i, parameters in enumerate(model[resource][op]['request_params']):
                                                if type(parameters) is list:
                                                    for j, parameter in enumerate(parameters):
                                                        if 'path_param' in parameter and parameter['name'] != param and parameter['name']:
                                                            print('Warning found duplicate path parameters: \'' + parameter['name'] + '\' and \'' + parameter['name'] + '\' in resource \'' + resource\
                                                                + '\' and operation \'' + operation + '\'')
                                                            print('A random path parameter will be chosen')
                                                            duplicate_path_param = True
                                                            break
                                                        elif 'path_param' in parameter and parameter['name'] == param:
                                                            duplicate_path_param = True
                                                        elif 'path_param' in parameter and not parameter['name']:
                                                            duplicate_path_param = True
                                                            model[resource][op]['request_params'][i][j]['name'] = param
                                            for i, parameter in enumerate(model[resource][op]['request_params']):
                                                if type(parameter) is not list:
                                                    if 'path_param' in parameter and parameter['name'] != param and parameter['name']:
                                                        print('Warning found duplicate path parameters: \'' + parameter['name'] + '\' and \'' + parameter['name'] + '\' in resource \'' + resource\
                                                            + '\' and operation \'' + operation + '\'')
                                                        print('A random path parameter will be chosen')
                                                        duplicate_path_param = True
                                                    elif 'path_param' in parameter and parameter['name'] == param:
                                                        duplicate_path_param = True
                                                    elif 'path_param' in parameter and not parameter['name']:
                                                        duplicate_path_param = True
                                                        model[resource][op]['request_params'][i]['name'] = param
                                            if duplicate_path_param == False:
                                                model[resource][op]['request_params'].append({'name':param,'path_param':True})
                                    else:
                                        plain_params = detect_parameters(step['sentence'],resource_names)
                                        for param in plain_params:
                                            if not param in plain_parameter_occurence[resource][op]['request']:
                                                plain_parameter_occurence[resource][op]['request'][param] = {'count':1}
                                            else:
                                                plain_parameter_occurence[resource][op]['request'][param]['count'] += 1
                                            duplicate = False
                                            for p in model[resource][op]['request_params']:
                                                if not type(p) is list and 'name' in p:
                                                    if p['name'] == param:
                                                        duplicate = True
                                            if duplicate == False:
                                                model[resource][op]['request_params'].append({'name':param})
                        for op in when_ops:
                            # roles apply to either the entire resource or a specific scenario
                            for resource_role in resource_roles:
                                scope = {'operation':op, 'resource':resource}
                                if not scope in model[resource]['resource_roles'][resource_role]:
                                    model[resource]['resource_roles'][resource_role].append(scope)
                            for scenario_role in scenario_roles:
                                if scenario_role not in model[resource]['resource_roles']:
                                    model[resource]['resource_roles'][scenario_role] = []
                                scope = {'operation':op, 'resource':resource}
                                if scope not in model[resource]['resource_roles'][scenario_role]:
                                    model[resource]['resource_roles'][scenario_role].append(scope)
                    elif steps_type == 'Then':
                        #preprocessing makes sure that 'when' is analyzed before 'then', so by now we know the scenario operations
                        steps_message_type = ''
                        steps_message_text = ''
                        for step in steps:
                            message = detect_messages(step['sentence'])
                            if message['text']:
                                steps_message_type = message['type']
                                steps_message_text = message['text']
                                if message['type'] == 'default':
                                    print('Warning, could not figure a status code from the phrase \'' + message['text'] + '\', the \'default\' be used as a status code')
                        if not steps_message_type:
                            steps_message_type = 'default'
                        for op in when_ops:
                            if steps_message_type not in model[resource][op]['responses']: # unique response types
                                model[resource][op]['responses'][steps_message_type] = {'params':[],'links':[],'param_examples':[],'message':steps_message_text}
                            if not steps_message_type+'_count' in operation_occurence[resource][op]:
                                operation_occurence[resource][op][steps_message_type+'_count'] = 1
                            else:
                                operation_occurence[resource][op][steps_message_type+'_count'] += 1
                        for step in steps:
                            message = detect_messages(step['sentence'])
                            if not message['text']:
                                if step['data_table']:
                                    processed_data_table = analyze_data_table(step['data_table'],None)
                                    if step['sentence'].endswith(':'): # sentence that describes a model
                                        model_name = detect_model_name(step['sentence'])
                                        if not model_name:
                                            model_name = 'Uknown_Model'
                                        for op in when_ops:
                                            if not any(model_name == param['name'] for param in model[resource][op]['responses'][steps_message_type]['params']):
                                                if processed_data_table['example']:
                                                    model[resource][op]['responses'][steps_message_type]['params'].append({'name':model_name, 'model_type':True, 'properties':processed_data_table['domain'], 'example':processed_data_table['example']})
                                                else:
                                                    model[resource][op]['responses'][steps_message_type]['params'].append({'name':model_name, 'model_type':True, 'properties':processed_data_table['domain']})
                                            if op not in plain_parameter_occurence[resource].keys():                                     
                                                plain_parameter_occurence[resource][op] = {steps_message_type+'_params':{}}                   
                                            else:
                                                if steps_message_type+'_params' not in plain_parameter_occurence[resource][op]:
                                                    plain_parameter_occurence[resource][op][steps_message_type+'_params'] = {}   
                                            if not model_name in plain_parameter_occurence[resource][op][steps_message_type+'_params']:
                                                plain_parameter_occurence[resource][op][steps_message_type+'_params'][model_name] = {'count':1}
                                            else:
                                                plain_parameter_occurence[resource][op][steps_message_type+'_params'][model_name]['count'] += 1 
                                    else:
                                        for op in when_ops:                                    
                                            if op not in table_parameter_occurence[resource]:            
                                                table_parameter_occurence[resource][op] = {steps_message_type+'_params':{}}                   
                                            elif (steps_message_type+'_params') not in table_parameter_occurence[resource][op]: 
                                                table_parameter_occurence[resource][op][steps_message_type+'_params'] = {}                                 
                                            if not processed_data_table['domain'] in model[resource][op]['responses'][steps_message_type]['params']:
                                                model[resource][op]['responses'][steps_message_type]['params'].append(processed_data_table['domain'])
                                                if processed_data_table['example']:
                                                    model[resource][op]['responses'][steps_message_type]['param_examples'].append(processed_data_table['example'])
                                                for param in processed_data_table['domain']:
                                                    if not param['name'] in table_parameter_occurence[resource][op][steps_message_type+'_params']:
                                                        table_parameter_occurence[resource][op][steps_message_type+'_params'][param['name']] = {'count':1}
                                                    else:
                                                        table_parameter_occurence[resource][op][steps_message_type+'_params'][param['name']]['count'] += 1
                                else:
                                    links = detect_operations(step['sentence'],resource_names,steps_type)
                                    if links:
                                        for op in when_ops:
                                            state = ''
                                            if steps_message_text:
                                                state = step['scenario_name'] + ' ('  + steps_message_text + ')'
                                            else:
                                                state = step['scenario_name'] + ' (' + 'default response' + ')'
                                            if resource+' '+state not in state_graph:
                                                state_graph[resource+' '+state] = {'operation':op,'resource':resource,'links':[],\
                                                'http_state':step['scenario_name'] + ' ('  + steps_message_type + ')','natural_state':state,\
                                                'code':steps_message_type} 
                                            for link in links:  
                                                if not ({'operation':link['operation'],'resource':link['path']} in model[resource][op]['responses'][steps_message_type]['links']):
                                                    flag = False
                                                    for oper in model[resource]:
                                                        if 'responses' in model[resource][oper]:
                                                            for message_type in model[resource][oper]['responses']:
                                                                if {'operation':link['operation'],'resource':link['path']} in model[resource][oper]['responses'][message_type]['links']: # make sure the link between the resources is not already registered
                                                                    flag = True
                                                    if flag == False:
                                                        resource_graph[resource].append({'operation':link['natural_verb'],'resource':link['natural_resource_name']})
                                                if link not in state_graph[resource+' '+state]['links']:
                                                    state_graph[resource+' '+state]['links'].append(link)                                           
                                                model[resource][op]['responses'][steps_message_type]['links'].append({'operation':link['operation'],'resource':link['path']})
                                    else:                                
                                        params = detect_parameters(step['sentence'],resource_names)
                                        for op in when_ops:
                                            if op not in plain_parameter_occurence[resource]:                                     
                                                plain_parameter_occurence[resource][op] = {steps_message_type+'_params':{}}                   
                                            elif steps_message_type+'_params' not in plain_parameter_occurence[resource][op]:
                                                plain_parameter_occurence[resource][op][steps_message_type+'_params'] = {}                           
                                            for param in params:
                                                if not param in plain_parameter_occurence[resource][op][steps_message_type+'_params']:
                                                    plain_parameter_occurence[resource][op][steps_message_type+'_params'][param] = {'count':1}
                                                else:
                                                    plain_parameter_occurence[resource][op][steps_message_type+'_params'][param]['count'] += 1
                                                duplicate = False
                                                for p in model[resource][op]['responses'][steps_message_type]['params']:
                                                    if not type(p) is list and 'name' in p:
                                                        if p['name'] == param:
                                                            duplicate = True
                                                if duplicate == False:
                                                    model[resource][op]['responses'][steps_message_type]['params'].append({'name':param})
                    elif steps_type == 'Scenario Description':
                        if not model[resource][op]['description']:
                            model[resource][op]['description'] = '- ' + steps
                        else:
                            model[resource][op]['description'] += "\n- " + steps
    model = set_parameter_requirements(table_parameter_occurence,operation_occurence,model,0)
    model = set_parameter_requirements(plain_parameter_occurence,operation_occurence,model,1)
    model = set_parameter_requirements(put_non_body_par_occurence,operation_occurence,model,2)

    # there can't be spaces in URIs so we replace spaces in resource names with underscores
    for resource in resources:
        compact_resource_name = '_'.join(resource.split(' '))
        model[compact_resource_name] = model.pop(resource)

    return {'model':model,'resource_graph':resource_graph, 'state_graph':state_graph}

def detect_other_resources(sentence,resource,resource_names):
    for resource_name in resource_names:
        if resource_name != resource and resource_name != p.plural(resource):
            if re.search(resource_name, sentence):
                return resource_name

def detect_roles(sentence,resource_names,resource_roles):
    tagged_tokens = TextBlob(sentence).tags
    roles = []
    has_resources = any(re.search(r'\b' + resource_name + r'\b',sentence) for resource_name in resource_names)
    if not has_resources: 
        for tagged_token in tagged_tokens:
            if tagged_token[0] not in resource_roles:
                if tagged_token[1][0:2] == 'NN':
                    roles.append(tagged_token[0])
    elif has_resources:
        if detect_http_verbs(sentence):
            for tagged_token in tagged_tokens:
                if tagged_token[0] not in resource_roles:
                    if tagged_token[1][0:2] == 'NN':
                        roles.append(tagged_token[0])
    return roles

def detect_messages(sentence):
    message = {'text':'','type':''}
    quoted_phrase = find_quoted_text(sentence)

    #assuming there is only one quoted phrase atm
    if quoted_phrase:
        plain_phrase = quoted_phrase[0].lower()
        message['text'] = quoted_phrase[0]
        language = TextBlob(message['text']).detect_language() # Requires internet connection since it uses the google translator API
        if ([text for text in L404[language] if re.search(text,plain_phrase)]):
            message['type'] = 'Not Found'
        elif ([text for text in L401[language] if re.search(text,plain_phrase)]):
            message['type'] = 'Unauthorized'
        elif ([text for text in L403[language] if re.search(text,plain_phrase)]):
            message['type'] = 'Forbidden'
        elif ([text for text in L400[language] if re.search(text,plain_phrase)]):
            message['type'] = 'Bad Request'
        elif ([text for text in L200[language] if re.search(text,plain_phrase)]):
            message['type'] = 'OK'
        elif ([text for text in L201[language] if re.search(text,plain_phrase)]):
            message['type'] = 'Created'
        elif ([text for text in L202 if re.search(text,plain_phrase)]):
            message['type'] = 'Accepted'
        if not message['type']:
            message['type'] = 'default'
    return message

def detect_operations(sentence,resource_names,step_type):
    # by convention a when_step can have ops and params in the same sentence
    # on the contrary a then_step sentence that describes operations is hateoas and cannot also have params
    tokenized_sentence = nltk.word_tokenize(sentence)
    if step_type == 'When':
        detected_verbs = detect_http_verbs(sentence)
        operations = []
        for detected_verb in detected_verbs:
            operations.append(detected_verb['http'])
        return operations
    elif step_type == 'Then':
        detected_verbs = detect_http_verbs(sentence)
        for resource_name in resource_names:
            resource_name_parts = resource_name.split(' ')
            if len(resource_name_parts) > 1:
                if all(re.search(r'\b' + part + r'\b',sentence) for part in resource_name_parts):
                    hateoas_links = []
                    for detected_verb in detected_verbs:
                        compact_resource_name = '_'.join(resource_name.split(' ')) # paths can't have spaces in them
                        hateoas_links.append({'natural_resource_name':resource_name,'path':compact_resource_name,'natural_verb':detected_verb['natural'], 'operation':detected_verb['http']})
                    return hateoas_links
        for resource_name in resource_names:
            if re.search(r'\b' + resource_name + r'\b',sentence) and not any(verb['natural'] == resource_name for verb in detected_verbs):
                hateoas_links = []
                for detected_verb in detected_verbs:
                    compact_resource_name = '_'.join(resource_name.split(' ')) # paths can't have spaces in them
                    hateoas_links.append({'natural_resource_name':resource_name,'path':compact_resource_name,'natural_verb':detected_verb['natural'], 'operation':detected_verb['http']})
                return hateoas_links

def detect_parameters(sentence,resource_names):
    tagged_tokens = TextBlob(sentence).tags
    parameters = []
    for tagged_token in tagged_tokens: # a tagged token has the word at position 0 and the tag at position 1
        if tagged_token[1][0:2] == 'NN':
            if not tagged_token[0] in resource_names:
                parameters.append(tagged_token[0])

    return parameters

def detect_model_name(sentence):
    tagged_tokens = TextBlob(sentence).tags
    detected_nouns = []
    for tagged_token in tagged_tokens: # a tagged token has the word at position 0 and the tag at position 1
        if tagged_token[1][0:2] == 'NN':
            detected_nouns.append(tagged_token[0])
    return "_".join(detected_nouns)

def detect_http_verbs(sentence):
    tagged_tokens = TextBlob(sentence).tags
    verbs = []
    for tagged_token in tagged_tokens: # a tagged token has the word at position 0 and the tag at position 1
        if tagged_token[1][0:2] == 'VB':           
            for http_verb, language_verbs in HTTP_verbs.items():
                if tagged_token[0] in language_verbs:
                   verbs.append({'http': http_verb, 'natural': tagged_token[0]})

    return verbs

def plural_extend(words_in_dict_keys):
    return list(words_in_dict_keys) + [p.plural(word) for word in list(words_in_dict_keys)]

def analyze_data_table(table,operation):
    domain = []
    example = {}
    numrows = len(table)
    numcolumns = len(table[0])   
    most_left_column = []
    second_row = []
    second_column = []
    third_row = []
    third_column = []
    fourth_row = []
    fourth_column  = []
    top_row = table[0]
    for row in table:
        most_left_column.append(row[0])
    if not any(is_data_type(cell) for cell in top_row): #domain is on the top row
        if numrows == 2:
            second_row = table[1]
        elif numrows == 3:
            second_row = table[1]
            third_row = table[2]
        elif numrows == 4:
            second_row = table[1]
            third_row = table[2]
            fourth_row = table[3]
        if second_row:
            if all(is_data_type(cell) for cell in second_row):
                for top_row_cell,second_row_cell in zip(top_row,second_row):
                    cell_type = figure_cell_type(top_row_cell,second_row_cell,operation)
                    domain.append(cell_type['domain'])
                    if cell_type['example']:
                        example[cell_type['example']['name']] = cell_type['example']['value']
                if third_row:
                    for top_row_cell,third_row_cell in zip(top_row,third_row):
                        domain = detect_min_max_cell(top_row_cell,third_row_cell,domain)
                if fourth_row:
                    for top_row_cell,fourth_row_cell in zip(top_row,fourth_row):
                        domain = detect_min_max_cell(top_row_cell,fourth_row_cell,domain)
            else:
                sys.exit("Incorrect table format")
        else:
            for cell in top_row:
                domain.append({'name':cell,'type':'string'})
    elif not any(is_data_type(cell) for cell in most_left_column): #domain is on the most left column  
        if numcolumns == 1:
            if len(table[0]) > 1:
                second_column.append(table[0][1])
        elif numcolumns == 2:
            for row in table:
                second_column.append(row[1])
        elif numcolumns == 3:
            for row in table:
                second_column.append(row[1])
                third_column.append(row[2])
        elif numcolumns == 4:
            for row in table:
                second_column.append(row[1])
                third_column.append(row[2])
                fourth_column.append(row[3])       
        if second_column:
            if all(is_data_type(cell) for cell in second_column):
                for most_left_column_cell,second_column_cell in zip(most_left_column,second_column):
                    cell_type = figure_cell_type(most_left_column_cell,second_column_cell,operation)
                    domain.append(cell_type['domain'])
                    if cell_type['example']:
                        example[cell_type['example']['name']] = cell_type['example']['value']
            else:
                sys.exit("Incorrect table format")
            if third_column:
                for most_left_column_cell,third_column_cell in zip(most_left_column,third_column):
                    domain = detect_min_max_cell(most_left_column_cell,third_column_cell,domain)
            if fourth_column:
                for most_left_column_cell,fourth_column_cell in zip(most_left_column,fourth_column):
                    domain = detect_min_max_cell(most_left_column_cell,fourth_column_cell,domain) 
        else:
            for cell in most_left_column:
                domain.append({'name':cell,'type':'string'})
    else:
        sys.exit("Incorrect table format")
    
    return {'domain':domain,'example':example}

def figure_cell_type(parameter_cell,type_cell,operation):
    domain_type = type_of_value(type_cell)
    domain = {}
    example = {}
    if domain_type is float:
        domain = {'name':parameter_cell,'type':'number','format':'float'}
        example = {'name':parameter_cell, 'value':ast.literal_eval(type_cell)}
    elif domain_type is int:
        domain = {'name':parameter_cell,'type':'integer','format':'int32'}
        example = {'name':parameter_cell, 'value':ast.literal_eval(type_cell)}
    elif domain_type is bool:
        domain = {'name':parameter_cell,'type':'boolean'}
        if type_cell == 'true':
            example = {'name':parameter_cell, 'value':True}
        else:
            example = {'name':parameter_cell, 'value':False}
    elif is_array(type_cell):                   
        content = eval(type_cell)
        values = []
        collection_format = 'unknown'
        types = None
        example = {'name':parameter_cell, 'value':[]}
        # print(content[0], type(content[0]))
        if type(content[0]) is list:
            types = list
            for item in content:
                example["value"].append(item)
        else:
            # if ',' in content:
            #     collection_format = 'csv'
            #     values = content.split(',')                            
            # elif ' ' in content:
            #     collection_format = 'ssv'
            #     values = content.split(' ')
            # elif '\t' in content:
            #     collection_format = 'tsv'
            #     values = content.split('\t')
            # elif '|' in content:
            #     collection_format = 'pipes'
            #     values = content.split('|')
            # else:
            #     collection_format = 'unknown'
            #     values = content
            types = type_of_value(content[0]) # assuming that user did not submit an array with different data types
        if types is float:
            domain = {'name':parameter_cell,'type':'array','collectionFormat':collection_format,'items':{'type':'number','format':'float'}}
            for item in content:
                example['value'].append(ast.literal_eval(item))
        elif types is int:
            domain = {'name':parameter_cell,'type':'array','collectionFormat':collection_format,'items':{'type':'integer','format':'int32'}}
            for item in content:
                example['value'].append(ast.literal_eval(item))
        elif types is bool:
            domain = {'name':parameter_cell,'type':'array','collectionFormat':collection_format,'items':{'type':'boolean'}}
            for item in content:
                if item == 'true':
                    example['value'].append(True)
                else:
                    example['value'].append(False)
        elif types is list:
            domain = {'name':parameter_cell,'type':'array','collectionFormat':collection_format,'items':{'type':'array', 'items' : {'type' : 'string'}}}
        else:
            domain = {'name':parameter_cell,'type':'array','collectionFormat':collection_format,'items':{'type':'string'}}
            for item in content:
                # print(item[1:-1])
                example['value'].append(item) # strip quotes
    else:
        if is_password(type_cell) and\
        (operation != 'delete' and operation != 'get'):
            domain = {'name':parameter_cell,'type':'string','format':'password'}
        elif not is_quoted_text(type_cell):
            if domain_type == 'file':
                domain = {'name':parameter_cell,'type':'file'}
            elif is_date(type_cell) and (operation != 'delete' or operation != 'get' or operation =='put_non_body'):
                if ':' in type_cell:
                    domain = {'name':parameter_cell,'type':'string','format':'date-time'} # "2017-02-19T20:51:09.0+02:00Z"                    
                else:
                    domain = {'name':parameter_cell,'type':'string','format':'date'} # "2017-02-19"
                example =  {'name':parameter_cell, 'value': type_cell}
        else:
            domain = {'name':parameter_cell,'type':'string'}
            example = {'name':parameter_cell, 'value':type_cell[1:-1]} #strip quotes
    return {'domain':domain, 'example':example}

def detect_numbers(text):
    numbers = []
    for word in text.split(' '):
        if type_of_value(word) is float or type_of_value(word) is int:
            numbers.append(ast.literal_eval(word))
    return numbers

def detect_min_max_cell(parameter_cell,type_cell,domain):
    numbers = detect_numbers(type_cell)
    if numbers:
        if len(numbers) == 2:
            minimum = numbers[0]
            maximum = numbers[1]
            if numbers[0] > numbers[1]:
                minimum = numbers[1]
                maximum = numbers[0]                                
            for i,param in enumerate(domain):
                if param['name'] == parameter_cell:
                    if param['type'] == 'string' or param['type'] == 'password':
                        domain[i]['minLength'] = minimum
                        domain[i]['maxLength'] = maximum
                    elif param['type'] == 'number' or param['type'] == 'integer':
                        domain[i]['minimum'] = minimum
                        domain[i]['maximum'] = maximum
                    elif param['type'] == 'array':
                        domain[i]['minItems'] = minimum
                        domain[i]['maxItems'] = maximum 
        elif len(numbers) == 1:
            if detect_phrase(type_cell,maximum_phrases):
                for i,param in enumerate(domain):
                    if param['name'] == parameter_cell:
                        if param['type'] == 'string' or param['type'] == 'password':
                            domain[i]['maxLength'] = numbers[0]
                        elif param['type'] == 'number' or param['type'] == 'integer':
                            domain[i]['maximum'] = numbers[0]
                        elif param['type'] == 'array':
                            domain[i]['maxItems'] = numbers[0]
            elif detect_phrase(type_cell,minimum_phrases):
                for i,param in enumerate(domain):
                    if param['name'] == parameter_cell:
                        if param['type'] == 'string' or param['type'] == 'password':
                            domain[i]['minLength'] = numbers[0]
                        elif param['type'] == 'number' or param['type'] == 'integer':
                            domain[i]['minimum'] = numbers[0]
                        elif param['type'] == 'array':
                            domain[i]['minItems'] = numbers[0]
    else:
        for i,param in enumerate(domain):
            if param['name'] == parameter_cell:
                if type_cell:
                    domain[i]['description'] = type_cell
    return domain

def set_parameter_requirements(param_occurences,operation_occurence,model,mode):
    #mode 0 = data table parameters
    #mode 1 = plain parameters
    #mode 2 = put non_body parameters
    if mode == 0:
        for resource,operations in param_occurences.items():
            for operation, occurences in operations.items():
                # set all parameters as not required
                for i, params in enumerate(model[resource][operation]['request_params']):
                    if type(params) is list:
                        for j, param in enumerate(params):
                            if not 'required' in model[resource][operation]['request_params'][i][j]:
                                model[resource][operation]['request_params'][i][j]['required'] = False
                for message_type in model[resource][operation]['responses']:
                    for i, params in enumerate(model[resource][operation]['responses'][message_type]['params']):
                        if type(params) is list:
                            for j, param in enumerate(params):
                                if not 'required' in model[resource][operation]['responses'][message_type]['params'][i][j]:
                                    model[resource][operation]['responses'][message_type]['params'][i][j]['required'] = False
                # find which are required
            for operation, occurences in operations.items():
                if 'request' in occurences:
                    for this_param_name in occurences['request'].keys():
                        if param_occurences[resource][operation]['request'][this_param_name]['count'] == operation_occurence[resource][operation]['request_count']:
                            for i, params in enumerate(model[resource][operation]['request_params']):
                                if type(params) is list:
                                    for j, param in enumerate(params):
                                        if param['name'] == this_param_name and\
                                        (param_occurences[resource][operation]['request'][this_param_name]['count'] == operation_occurence[resource][operation]['request_count']):
                                            model[resource][operation]['request_params'][i][j]['required'] = True
                for message_type in model[resource][operation]['responses']:
                    if message_type+'_params' in occurences:
                        for this_param_name in occurences[message_type+'_params'].keys():
                            for i, params in enumerate(model[resource][operation]['responses'][message_type]['params']):
                                if type(params) is list:
                                    for j, param in enumerate(params):
                                        if param['name'] == this_param_name and \
                                        (param_occurences[resource][operation][message_type+'_params'][this_param_name]['count'] == operation_occurence[resource][operation][message_type+'_count']):
                                            model[resource][operation]['responses'][message_type]['params'][i][j]['required'] = True
    elif mode == 1:
        for resource,operations in param_occurences.items():
            for operation, occurences in operations.items():
                # set all parameters as not required
                for i, param in enumerate(model[resource][operation]['request_params']):
                    if type(param) is not list:
                        if not 'required' in model[resource][operation]['request_params'][i]:
                            model[resource][operation]['request_params'][i]['required'] = False
                for message_type in model[resource][operation]['responses']:
                    for i, param in enumerate(model[resource][operation]['responses'][message_type]['params']):
                        if type(param) is not list:
                            if not 'required' in model[resource][operation]['responses'][message_type]['params'][i]:
                                model[resource][operation]['responses'][message_type]['params'][i]['required'] = False
                # find which are required
            for operation, occurences in operations.items():
                if 'request' in occurences:
                    for this_param_name in occurences['request'].keys():
                        if param_occurences[resource][operation]['request'][this_param_name]['count'] == operation_occurence[resource][operation]['request_count']:
                            for i, param in enumerate(model[resource][operation]['request_params']):
                                if type(param) is not list:
                                    if param['name'] == this_param_name and\
                                    (param_occurences[resource][operation]['request'][this_param_name]['count'] == operation_occurence[resource][operation]['request_count']):
                                        model[resource][operation]['request_params'][i]['required'] = True
                for message_type in model[resource][operation]['responses']:
                    if message_type+'_params' in occurences:
                        for this_param_name in occurences[message_type+'_params'].keys():
                            for i, param in enumerate(model[resource][operation]['responses'][message_type]['params']):
                                if type(param) is not list:
                                    if param['name'] == this_param_name and \
                                    (param_occurences[resource][operation][message_type+'_params'][this_param_name]['count'] == operation_occurence[resource][operation][message_type+'_count']):
                                        model[resource][operation]['responses'][message_type]['params'][i]['required'] = True
    elif mode == 2:
        for resource,operations in param_occurences.items():
            for operation, occurences in operations.items():
                # set all parameters as not required
                if operation == 'put':
                    for i, params in enumerate(model[resource][operation]['non_body_params']):
                        if type(params) is list:
                            for j, param in enumerate(params):
                                if not 'required' in model[resource][operation]['non_body_params'][i][j]:
                                    model[resource][operation]['non_body_params'][i][j]['required'] = False
            for operation, occurences in operations.items():
                # find which are required
                if operation == 'put':
                    # find which are required
                    if 'request' in occurences:
                        for this_param_name in occurences['request'].keys():
                            if param_occurences[resource][operation]['request'][this_param_name]['count'] == operation_occurence[resource][operation]['request_count']:
                                for i, params in enumerate(model[resource][operation]['request_params']):
                                    if type(params) is list:
                                        for j, param in enumerate(params):
                                            if param['name'] == this_param_name and\
                                            (param_occurences[resource][operation]['request'][this_param_name]['count'] == operation_occurence[resource][operation]['request_count']):
                                                model[resource][operation]['non_body_params'][i][j]['required'] = True

    return model

def detect_phrase(text,phrases):
    for phrase in phrases:
        if re.search(phrase,text):
            return True
    return False

def type_of_value(var):
    try:
        return type(ast.literal_eval(var))
    except Exception:
        if var == 'true' or var == 'false':
            return bool
        elif var == 'file':
            return 'file'
        else:
            return str

def is_date(string):
    try: 
        parse(string)
        return True
    except ValueError:
        return False

def find_quoted_text(string):
    return re.findall(r'\"(.+?)\"',string) or re.findall(r'\'(.+?)\'',string)

def is_array(string):
    return re.findall(r'\[(.+?)\]',string)

def is_quoted_text(string):
    if re.findall(r'\"(.+?)\"',string) or re.findall(r'\'(.+?)\'',string):
        return True
    else:
        return False

def is_data_type(string):
    if is_quoted_text(string):
        return True
    value_type = type_of_value(string)
    if (value_type is float) or (value_type is int) or (value_type is bool) or (value_type is list) or (value_type == 'file')\
     or is_array(string) or is_password(string):
        return True
    elif is_date(string):
        return True
    else:
        return False

def is_password(string):
    for character in string:
        if character != '*':
            return False    
    return True



from .utils import *
import json
import os
import traceback

def gherkin2oas(resources_folder=None, oas_title=None, oas_description=None, oas_security=None, oas_version=None, oas_tos=None, oas_host=None, oas_basepath=None, oas_schemes=None, oas_produces=None):
    try:
        if not os.path.isdir(resources_folder + '/output_files'):
            os.makedirs(resources_folder + '/output_files')
        if not os.path.isdir(resources_folder + '/output_files/swagger-oas2'):
            os.makedirs(resources_folder + '/output_files/swagger-oas2')
        if not os.path.isdir(resources_folder + '/output_files/oas3'):
            os.makedirs(resources_folder + '/output_files/oas3')
    except:
        print(traceback.format_exc())
    resources = preprocessor.main(resources_folder)
    resource_names = nlp.plural_extend(resources)
    nlp_model = nlp.resource_analysis(resources,resource_names)
    model = nlp_model['model']
    resource_graph = nlp_model['resource_graph']
    state_graph = nlp_model['state_graph']
    oas_schema = formatter.generate_swagger(model, oas_security)
    oas_schema['swagger'] = '2.0'
    oas_schema['info'] = {  "title":            oas_title,
                            "description":      oas_description,
                            "version":          oas_version,
                            "termsOfService":   oas_tos
    }
    oas_schema['host'] = oas_host
    oas_schema['schemes'] = oas_schemes 
    oas_schema['basePath'] = oas_basepath
    oas_schema['produces'] = oas_produces

    with open(resources_folder + '/output_files/swagger-oas2/swagger.json', 'w') as outfile:
        json.dump(oas_schema, outfile, indent=4)


from gherkin.parser import Parser
import os
from os.path import basename
from collections import OrderedDict

# General comment: python dicts not perserving order matters here. Order is required.

def main(resources_path):
    parser = Parser()

    nlp_ready_resources = {}

    for root, dirs, files in os.walk(resources_path):
        for file_name in files:
            if file_name.endswith('.resource'):
                resource = os.path.splitext(basename(file_name))[0]
                resource = ' '.join(resource.split('-'))
                resource = ' '.join(resource.split('_'))
                parsed_resource_file = parser.parse(os.path.join(root, file_name))
                nlp_ready_resources[resource] = {}
                for child in parsed_resource_file['feature']['children']:
                    if child['type'] == 'Background':
                        nlp_ready_resources[resource]['background'] = {}
                        nlp_ready_resources[resource]['background']['Given'] = []
                        for step in child['steps']:
                            sentence = step['keyword'] + step['text']
                            nlp_ready_resources[resource]['background']['Given'].append({'sentence':sentence})
                    elif child['type'] == 'Scenario':
                        ordered_step_types = OrderedDict({'Given':[],'When':[],'Then':[]})
                        ordered_step_types.move_to_end('When')
                        ordered_step_types.move_to_end('Then')
                        nlp_ready_resources[resource][child['name']] = ordered_step_types
                        in_step = ''
                        for step in child['steps']:
                            data_table = []
                            sentence = step['keyword'] + step['text']
                            if step['keyword'] == 'When ' or step['keyword'] == 'Then ' or step['keyword'] == 'Given ': 
                                in_step = step['keyword'].strip() #note: there is a space here after the keyword                
                            if 'argument' in step:
                                if step['argument']['type'] == 'DataTable':
                                    data_table = parse_table(step)
                            if not in_step == 'Given':
                                nlp_ready_resources[resource][child['name']][in_step].append({'sentence':sentence,'data_table':data_table,'scenario_name':child['name']})
                                if 'description' in child:
                                    nlp_ready_resources[resource][child['name']]['Scenario Description'] = child['description']
                            else:
                                nlp_ready_resources[resource][child['name']][in_step].append({'sentence':sentence,'scenario_name':child['name']})
    return nlp_ready_resources

def parse_table(step):
    data_table = []
    for row in step['argument']['rows']:
        r = []
        for cell in row['cells']:
            r.append(cell['value'])
        data_table.append(r)
    return data_table


import re
import sys

http_codes = {'OK': '200', 'Created':'201', 'Accepted':'202', 'Bad Request': '400', 'Unauthorized': '401', 'Not Found': '404', 'Forbidden': '403', 'default': 'default'}
ids=[]

def generate_swagger(model, security):
    global ids
    paths_object = {}
    definitions_object = {}
    security_definitions_object = {}
    security_object = []
    all_paths = {}
    empty = {'get':{'request_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}}, 'description':''},
        'post':{'request_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}},'description':''},
        'put':{'request_params':[],'non_body_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}},'description':''},
        'delete':{'request_params':[],'request_examples':[],'responses':{'default':{'params':[],'links':[],'param_examples':[],'message':''}},'description':''}}
    path_hierarchy = {}
    roles = {}
    for resource,operations in model.items():
        path_hierarchy[resource] = model[resource].pop('resource_hierarchy')
        roles[resource] = model[resource].pop('resource_roles')
        # ---------------path seeds--------------------        
        collection_path = '/' + resource
        single_resource_path = ''
        get_query_path = False
        delete_query_path = False
        all_paths[resource] = {}
        # ---------------resource identifications--------------------
        for operation,op_model in operations.items():
            if op_model != empty[operation] and (operation != 'post'):        
                if operation == 'get' or operation == 'delete' or operation == 'put':   
                    flattened_params = list(flatten(op_model['request_params']))
                    if len(flattened_params) == 1:
                        param = flattened_params[0]
                        if not (operation == 'put' and 'model_type' in param): 
                            if param['name']:
                                single_resource_path = collection_path + '/{'+param['name']+'}'
                                all_paths[resource][operation] = single_resource_path 
                                if not single_resource_path in paths_object:
                                    paths_object[single_resource_path] = {}
                                if not operation in paths_object[single_resource_path]:
                                    paths_object[single_resource_path][operation] = {'parameters':[]}
                                paths_object[single_resource_path][operation]['parameters'].append(constr_path(param))
                    else:
                        if len(flattened_params) != 0:
                            all_paths[resource][operation] = collection_path
                            path_param = {}
                            for param in flattened_params:
                                if not (operation == 'put' and 'model_type' in param):  
                                    if 'path_param' in param and param['name']:
                                        path_param = param
                                        single_resource_path =  collection_path + '/{'+param['name']+'}'
                                        all_paths[resource][operation] =  single_resource_path
                                        if not single_resource_path in paths_object:
                                            paths_object[single_resource_path] = {}
                                        if not operation in paths_object[single_resource_path]:
                                            paths_object[single_resource_path][operation] = {'parameters':[]} 
                                        paths_object[single_resource_path][operation]['parameters'].append(constr_path(param))
        for operation,op_model in operations.items():
            if op_model != empty[operation] and (operation != 'post'):                 
                if len(op_model['request_params']) == 0:
                    if not single_resource_path:
                        single_resource_path = collection_path + '/{' + 'id}'
                        all_paths[resource][operation] = single_resource_path
                        paths_object[single_resource_path] = {}
                        paths_object[single_resource_path][operation] = {'parameters':[]}                     
                        paths_object[single_resource_path][operation]['parameters'].append({'name':'id','type':'integer','format':'int32','in':'path','required':True})
                    else: # assign a random, already found, single resource path
                        all_paths[resource][operation] = single_resource_path
                        if not operation in paths_object[single_resource_path]:
                            paths_object[single_resource_path][operation] = {'parameters':[]}
                        p = {}
                        for op,o_m in paths_object[single_resource_path].items():
                            for param in o_m['parameters']:
                                if param['in'] == 'path':
                                    p = param
                                    break
                        paths_object[single_resource_path][operation]['parameters'].append(p) 
        # ---------------queries--------------------             
        for operation,op_model in operations.items():
            if op_model != empty[operation] and (operation == 'get' or operation == 'delete'):        
                flattened_params = list(flatten(op_model['request_params']))
                if len(flattened_params) > 1:
                    if not all_paths[resource][operation]:
                        all_paths[resource][operation] = collection_path
                    if not collection_path in paths_object:
                        paths_object[collection_path] = {}
                    if not operation in paths_object[all_paths[resource][operation]]:
                        paths_object[all_paths[resource][operation]][operation] = {'parameters':[]}
                    for param in flattened_params:
                        if 'path_param' not in param and param['name']:
                            paths_object[all_paths[resource][operation]][operation]['parameters'].append(constr_query(param))            
                            if operation == 'get':
                                get_query_path = True
                            elif operation == 'delete':
                                delete_query_path = True                                                   
        for operation,op_model in operations.items():
            if op_model != empty[operation] and operation == 'put':
                if op_model['non_body_params']:
                    flattened_params = list(flatten(op_model['non_body_params']))
                    if not all_paths[resource][operation]:
                        all_paths[resource][operation] = collection_path
                    if not collection_path in paths_object:
                        paths_object[collection_path] = {}
                    if not operation in paths_object[all_paths[resource][operation]]:
                        paths_object[all_paths[resource][operation]][operation] = {'parameters':[]}
                    for param in flattened_params:
                        paths_object[all_paths[resource][operation]][operation]['parameters'].append(constr_query(param))                    
        # ---------------bodies--------------------   
        for operation,op_model in operations.items():
            if op_model != empty[operation] and (operation == 'post' or operation == 'put'):             
                object_name = ''
                if not operation in all_paths[resource]:
                    all_paths[resource][operation] = collection_path
                    this_path = collection_path
                else:
                    this_path = all_paths[resource][operation]
                if not this_path in paths_object:
                    paths_object[this_path] = {}
                if not operation in paths_object[this_path]:
                    paths_object[this_path][operation] = {'parameters':[]}
                if op_model['request_params']:
                    has_file = False
                    for param in op_model['request_params']:
                        if type(param) is list:
                            for domain_param in param:
                                if domain_param['type'] == 'file': # cannot have body AND formData parameters at the same time 
                                    paths_object[this_path][operation]['consumes'] = ["multipart/form-data","application/x-www-form-urlencoded"]
                                    paths_object[this_path][operation]['parameters'].append({'name':domain_param['name'],'in':'formData','type':domain_param['type'],'required':True}) 
                                    has_file = True
                    if not has_file:
                        object_name = resource + '_' + operation + '_request_body'
                        definitions_object[object_name] = {}
                        definitions_object[object_name]['type'] = 'object'
                        definitions_object[object_name]['properties'] = {}
                        if all(param['required'] == True for param in list(flatten(op_model['request_params']))):
                            paths_object[this_path][operation]['parameters'].append({'name':object_name,'in':'body','schema':{'$ref':'#/definitions/' + object_name},'required':True})
                        else:
                            paths_object[this_path][operation]['parameters'].append({'name':object_name,'in':'body','schema':{'$ref':'#/definitions/' + object_name}})
                        for param in op_model['request_params']:
                            if type(param) is list:
                                for domain_param in param:
                                    if 'required' in domain_param:
                                            if domain_param['required'] == True:
                                                if 'required' not in definitions_object[object_name].keys():
                                                    definitions_object[object_name]['required'] = [domain_param['name']]
                                                elif domain_param['name'] not in definitions_object[object_name]['required']:
                                                    definitions_object[object_name]['required'].append(domain_param['name'])  
                                    definitions_object[object_name]['properties'][domain_param['name']] = constr_schema(domain_param)              
                            elif 'model_type' in param:
                                model_definition_name = param['name']
                                if not model_definition_name in definitions_object:
                                    definitions_object[model_definition_name] = {}
                                    definitions_object[model_definition_name]['type'] = 'object'
                                    definitions_object[model_definition_name]['properties'] = {}
                                if 'required' in param:
                                    if param['required'] == True:
                                        if 'required' not in definitions_object[object_name].keys():
                                            definitions_object[object_name]['required'] = [param['name']]
                                        elif param['name'] not in definitions_object[object_name]['required']:
                                            definitions_object[object_name]['required'].append(param['name'])
                                definitions_object[object_name]['properties'][param['name']] = {'$ref':'#/definitions/' + param['name']}
                                for model_property in param['properties']:                                
                                    definitions_object[model_definition_name]['properties'][model_property['name']] = constr_schema(model_property)
                                    if param['example']:
                                        definitions_object[model_definition_name]['example'] = [param['example']]       
                            else:
                                if 'path_param' not in param:
                                    definitions_object[object_name]['properties'][param['name']] = {'type':'string'}
                        for example in op_model['request_examples']:
                            if not 'example' in definitions_object[object_name]:
                                definitions_object[object_name]['example'] = [example]
                            else:
                                definitions_object[object_name]['example'].append(example)
                elif not op_model['request_params'] and operation == 'post':
                    sys.exit('A unique post operation inside a resource file must ALWAYS specify at least one body parameter', resource, operation)
        for operation,op_model in operations.items():
            if op_model['description']:
                this_path = all_paths[resource][operation]
                if operation == 'post' or (operation == 'get' and get_query_path) or (operation == 'delete' and delete_query_path):
                    paths_object[this_path][operation]['description'] = op_model['description']
                else:
                   paths_object[this_path][operation]['description'] = op_model['description']
    # ---------------path hierarchies--------------------
    hierarchy_levels = [[]]
    total_paths_count = len(list(paths_object))
    completed_paths_count = 0
    new_paths = []
    done_resources = {}
    for resource,operations in model.items():
        done_resources[resource] = False
        if not path_hierarchy[resource]:
            done_resources[resource] = True
            for operation in operations:
                if operations[operation] != empty[operation]:
                    if not any(all_paths[resource][operation] == registered_path['path'] for registered_path in hierarchy_levels[0]):
                        path_params = []
                        if paths_object[all_paths[resource][operation]][operation]['parameters']:
                            for parameter in paths_object[all_paths[resource][operation]][operation]['parameters']:
                                if parameter['in'] == 'path':
                                    path_params.append(parameter)            
                        hierarchy_levels[0].append({'path':all_paths[resource][operation], 'path_params':path_params})
                        new_paths.append({'new_path':all_paths[resource][operation],'old_path':all_paths[resource][operation], 'path_params':[]})
                        completed_paths_count += 1
    # if not hierarchy_levels[0]:
    #     sys.exit('It seems that a path hierarchy was specified in every resource file. For path hierarchies to make sense, at least one resource file must have no hierarchies')
    if hierarchy_levels[0]:
        high_hierarchy_level = 0
        low_hierarchy_level = 1
        hierarchy_levels.append([])
        while completed_paths_count != len(list(paths_object)):
            for resource in model:
                if path_hierarchy[resource] and not done_resources[resource]:
                    inherited_path_params = []
                    if path_hierarchy[resource]['linking_parameter']:
                        found_link = False
                        for higher_hierarchy_path in hierarchy_levels[high_hierarchy_level]:
                            if re.search(r'\b' + path_hierarchy[resource]['hierarchy_resource'] + r'\b',higher_hierarchy_path['path']):
                                if higher_hierarchy_path['path_params']:
                                    for path_param in higher_hierarchy_path['path_params']:
                                        if path_param['name'] == path_hierarchy[resource]['linking_parameter'] and re.search(path_param['name'] + '}$', higher_hierarchy_path['path']):
                                            found_link = True
                                            done_resources[resource] = True
                                    if found_link == True:
                                        for path_param in higher_hierarchy_path['path_params']:
                                            if path_param not in inherited_path_params:
                                                inherited_path_params.append(path_param)
                                        for operation in model[resource]:
                                            if model[resource][operation] != empty[operation]:
                                                if (operation == 'get' or operation == 'delete' or operation == 'put'):
                                                    for parameter in paths_object[all_paths[resource][operation]][operation]['parameters']:
                                                        if parameter['in'] == 'path' and parameter not in inherited_path_params:
                                                            inherited_path_params.append(parameter)
                                                new_path = higher_hierarchy_path['path'] + all_paths[resource][operation]
                                                if not any(registered_path['path'] == new_path for registered_path in hierarchy_levels[low_hierarchy_level]):
                                                    new_paths.append({'new_path':new_path, 'old_path':all_paths[resource][operation],'path_params':inherited_path_params})
                                                    hierarchy_levels[low_hierarchy_level].append({'path':new_path, 'path_params':inherited_path_params})
                                                    completed_paths_count += 1
                                        break                       
            high_hierarchy_level += 1
            low_hierarchy_level += 1
            hierarchy_levels.append([])

        for path in new_paths:
            new_path = path['new_path']
            old_path = path['old_path']
            if old_path in paths_object:
                paths_object[new_path] = paths_object.pop(old_path)
                for parameter in path['path_params']:
                    for operation in paths_object[new_path]:
                        if parameter not in paths_object[new_path][operation]['parameters']:
                            paths_object[new_path][operation]['parameters'].append(parameter)
    

        for resource in model:
            for operation in model[resource]:
                if model[resource][operation] != empty[operation]:
                    for path in new_paths:
                        if all_paths[resource][operation] == path['old_path']:
                            all_paths[resource][operation] = path['new_path']
                            break
    # ---------------responses--------------------
    for resource,operations in model.items():
        for operation,op_model in operations.items():
            if op_model != empty[operation]:
                this_path = all_paths[resource][operation]
                object_name = ''
                for response_type in op_model['responses']:
                    if op_model['responses'][response_type]['params'] or op_model['responses'][response_type]['links'] or op_model['responses'][response_type]['message']:
                        status_message = op_model['responses'][response_type]['message']
                        if not status_message:
                            status_message = 'None'
                        # create definition object entry
                        if op_model['responses'][response_type]['params'] or op_model['responses'][response_type]['links']:
                            object_name = resource + '_' + operation + '_' + http_codes[response_type] + '_response'
                            definitions_object[object_name] = {}
                            definitions_object[object_name]['type'] = 'object'
                        # create path object entry
                        if not 'responses' in paths_object[this_path][operation]:
                            paths_object[this_path][operation]['responses'] = {}
                        if not response_type in paths_object[this_path][operation]['responses']:
                            if op_model['responses'][response_type]['params'] or op_model['responses'][response_type]['links']:
                                if response_type in http_codes:
                                    paths_object[this_path][operation]['responses'][http_codes[response_type]] = {'description':status_message,'schema':{'$ref':'#/definitions/' + object_name}}
                                else:
                                    paths_object[this_path][operation]['responses']['default'] = {'description':'None','schema':{'$ref':'#/definitions/' + object_name}} 
                            elif op_model['responses'][response_type]['message']:
                                if response_type in http_codes:
                                    paths_object[this_path][operation]['responses'][http_codes[response_type]] = {'description':status_message}
                                else:
                                    paths_object[this_path][operation]['responses']['default'] = {'description':'None'}
                    if op_model['responses'][response_type]['params']: 
                        if not 'properties' in definitions_object[object_name]:                 
                            definitions_object[object_name]['properties'] = {}     
                        for param in op_model['responses'][response_type]['params']:
                            if type(param) is list:
                                for domain_param in param:             
                                    if 'required' in domain_param:
                                        if domain_param['required'] == True:
                                            if 'required' not in definitions_object[object_name]:
                                                definitions_object[object_name]['required'] = [domain_param['name']]
                                            elif domain_param['name'] not in definitions_object[object_name]['required']:
                                                definitions_object[object_name]['required'].append(domain_param['name'])
                                    definitions_object[object_name]['properties'][domain_param['name']] = constr_schema(domain_param)
                            elif 'model_type' in param:
                                model_definition_name = param['name']
                                if not model_definition_name in definitions_object:
                                    definitions_object[model_definition_name] = {}
                                    definitions_object[model_definition_name]['type'] = 'object'
                                    definitions_object[model_definition_name]['properties'] = {}
                                if 'required' in param:
                                    if param['required'] == True:
                                        if 'required' not in definitions_object[object_name].keys():
                                            definitions_object[object_name]['required'] = [param['name']]
                                        elif param['name'] not in definitions_object[object_name]['required']:
                                            definitions_object[object_name]['required'].append(param['name'])
                                definitions_object[object_name]['properties'][param['name']] = {'$ref':'#/definitions/' + model_definition_name}
                                for model_property in param['properties']:
                                    definitions_object[model_definition_name]['properties'][model_property['name']] = constr_schema(model_property)
                                if param['example']:
                                    definitions_object[model_definition_name]['example'] = [param['example']]  
                            else:
                                definitions_object[object_name]['properties'][param['name']] = {'type':'string'}
                        for example in op_model['responses'][response_type]['param_examples']:
                            if not 'example' in definitions_object[object_name]:
                                definitions_object[object_name]['example'] = [example]
                            else:
                                definitions_object[object_name]['example'].append(example)   
                    if op_model['responses'][response_type]['links']:
                        if not 'x-links' in definitions_object[object_name]:
                            definitions_object[object_name]['x-links'] = []                         
                        for link in op_model['responses'][response_type]['links']:
                            if link['resource'] in all_paths:
                                if link['operation'] in all_paths[link['resource']]:
                                    hateoas_link = {'path':all_paths[link['resource']][link['operation']],'operation':link['operation']}
                                    if hateoas_link not in definitions_object[object_name]['x-links']:
                                        definitions_object[object_name]['x-links'].append(hateoas_link) 
                                else:
                                    print('No ' + link['operation'] + ' operation found in resource file ' + link['resource'] + ' but such a hateoas link was described in ' + resource + ' resource file')
                            else:
                                print('No ' + link['operation'] + ' operation found in resource file ' + link['resource'] + ' but such a hateoas link was described in ' + resource + ' resource file')
    # ---------------security--------------------
    found_paths = list(paths_object.keys())
    flag_0 = False
    if security == 'access_token':   
        for resource,operations in model.items():
            security_definitions_object = {"cookieAuth" : {"type": "apiKey", "in" : "cookie", "name" : "access_token"}}
            for role,scopes in roles[resource].items():
                for scope in scopes:
                    paths_object[all_paths[resource][scope['operation']]][scope['operation']]["security"] = [{"cookieAuth" : []}]
                    flag_0 = True 
        if flag_0 == False:
            security_object = [{"cookieAuth" : []}]
    else:
        for resource,operations in model.items():
            for role,scopes in roles[resource].items():
                security_definitions_object[role] = {'type':'oauth2','flow':'implicit',"authorizationUrl": "http://swagger.io/api/oauth/dialog",'scopes':{}}
                for scope in scopes:
                    security_definitions_object[role]['scopes'][scope['operation']+':'+all_paths[resource][scope['operation']]] = 'No description'
                    if not 'security' in paths_object[all_paths[resource][scope['operation']]][scope['operation']].keys():
                        paths_object[all_paths[resource][scope['operation']]][scope['operation']]['security'] = [{role:[scope['operation']+':'+all_paths[resource][scope['operation']]]}]
                    else:
                        flag = False
                        for i, security_requirement in enumerate(paths_object[all_paths[resource][scope['operation']]][scope['operation']]['security']):
                            if role in paths_object[all_paths[resource][scope['operation']]][scope['operation']]['security'][i].keys():
                                paths_object[all_paths[resource][scope['operation']]][scope['operation']]['security'][i][role].append(scope['operation']+':'+all_paths[resource][scope['operation']])
                                flag = True
                        if flag == False:
                            paths_object[all_paths[resource][scope['operation']]][scope['operation']]['security'] = [{role:[scope['operation']+':'+all_paths[resource][scope['operation']]]}]
    # ---------------do some cleaning--------------------                            
    empty_defs = []
    for definition, definition_object in definitions_object.items():
        if 'properties' in definition_object:
            if not definition_object['properties']:
                empty_defs.append(definition)
    for empty_def in empty_defs:
        silence = definitions_object.pop(empty_def)
    # ---------------create unique path ids and update paths and params everywhere in the specification--------------------  
    generic_id_param = -1
    new_paths = []
    old_paths = paths_object.keys()
    for path in old_paths:
        new_path = re.sub(r'([^/]+)/{id}', rep, path)
        for operation in paths_object[path]:
            for i,parameter in enumerate(paths_object[path][operation]['parameters']):
                if parameter['in'] == 'path' and parameter['name'] == 'id' and ids:
                    generic_id_param = i
                    for new_id in ids:
                        paths_object[path][operation]['parameters'].append({'type':'integer','name':new_id,'required':True,'format':'int32','in':'path'})
            if generic_id_param != -1:
                silence = paths_object[path][operation]['parameters'].pop(generic_id_param)
            generic_id_param = -1
        ids = []
        if path in paths_object:
            paths_object[new_path] = paths_object.pop(path)

    for definition in definitions_object:
        if 'x-links' in definitions_object[definition]:
            new_x_links = []
            for x_link in definitions_object[definition]['x-links']:
                new_x_links.append({'path':re.sub(r'([^/]+)/{id}', rep, x_link['path']),'operation':x_link['operation']})
            definitions_object[definition]['x-links'] = new_x_links

    # remove duplicate parameters from queries, if any
    for path in paths_object:
        for operation in paths_object[path]:
            if operation == "get":
                unique_parameter_list = []
                for parameter in paths_object[path][operation]['parameters']:
                    if parameter not in unique_parameter_list:
                        unique_parameter_list.append(parameter)
                paths_object[path][operation]['parameters'] = unique_parameter_list
            

    return {'paths':paths_object,'definitions':definitions_object,'securityDefinitions':security_definitions_object, 'security':security_object}

def flatten(container):
    for i in container:
        if isinstance(i, (list,tuple)):
            for j in flatten(i):
                yield j
        else:
            yield i

def constr_path(param):
    param_properties = {}
    for param_property_name,param_property_value in param.items():
        if param_property_name in ['name','type','format','maxItems','minItems','minimum','maximum','minLength','maxLength','description']:
            param_properties[param_property_name] = param_property_value
    param_properties['required'] = True
    param_properties['in'] = 'path'
    if 'type' not in param_properties:
        if param['name'] == 'id':
            param_properties['type'] = 'integer'
            param_properties['format'] = 'int32'
        else:
            param_properties['type'] = 'string'
    return param_properties

def constr_query(param):
    param_properties = {}
    for param_property_name,param_property_value in param.items():
        if param_property_name in ['name','type','format','maxItems','minItems','minimum','maximum','minLength','maxLength','required','description']:
            param_properties[param_property_name] = param_property_value
    param_properties['in'] = 'query'
    if 'type' not in param_properties:
        if param['name'] == 'id':
            param_properties['type'] = 'integer'
            param_properties['format'] = 'int32'
        else:
            param_properties['type'] = 'string'
    return param_properties

def constr_schema(param):
    param_properties = {}
    for param_property_name,param_property_value in param.items():
        if param_property_name in ['type','format','maxItems','minItems','minimum','maximum','minLength','maxLength','items','description']:
            if param_property_name == 'type' and param_property_value == 'file':
                print('File type in response is bugged, see https://github.com/OAI/OpenAPI-Specification/issues/260')
            else:
                param_properties[param_property_name] = param_property_value
    return param_properties 

def rep(m):
    global ids
    ids.append('{}_id'.format(m.group(1)))
    return '{0}/{{{0}_id}}'.format(m.group(1))





import setuptools
from setuptools import _install_setup_requires

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gherkin2oas", # Replace with your own username
    version="1.0.0",
    author="Tasos Dimanidis",
    author_email="dhmtasos@gmail.com",
    description="A converter from Gherkin to OAS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TasosDhm/gherkin2oas",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
    install_requires=['nltk', 'TextBlob', 'inflect', 'python-dateutil', 'gherkin-official>=4.1.3']
)